package tsp.csc4509.blockfile;

import static tsp.csc4509.common.Log.GEN;
import java.io.IOException;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileChannel.MapMode;
import java.nio.channels.ReadableByteChannel;

/**
 * Classe qui décrit un bloc de données mappé sur le fichier à recevoir ou
 * copier.
 * 
 * @author Eric Lallet.
 *
 */
public class Block implements AutoCloseable {
	/**
	 * buffer mappé sur la zone [offset, offset+size[ du fichier relié au canal
	 * fcout.
	 */
	private MappedByteBuffer data;

	/**
	 * offset du premier octet mappé par le buffer data sur le fichier relié au
	 * canal fcout.
	 */
	private long offset;

	/**
	 * taille de la zone mappée par le buffer data sur le fichier relié au canal
	 * fcout, et donc aussi la taille du buffer data.
	 */
	private int size;

	/**
	 * canal relié au fichier où il faut écrire les données reçues.
	 */
	private FileChannel fcout;

	/**
	 * Constructeur de la classe Block qui initialise les principaux attributs. Pour
	 * éviter d'occuper trop de mémoire, le buffer data n'est pas mappé à la
	 * construction, mais est mappé au vol lors de son premier usage dans la méthode
	 * read().
	 * 
	 * @param offset offset du premier octet mappé par le buffer data sur le fichier
	 *               relié au canal fcout.
	 * @param size   taille de la zone mappée par le buffer data sur le fichier
	 *               relié au canal fcout, et donc, aussi la taille du buffer data.
	 * @param fcout  canal relié au fichier où il faut écrire les données reçues.
	 */
	public Block(final long offset, final int size, final FileChannel fcout) {
		this.offset = offset;
		this.size = size;
		this.fcout = fcout;
	}

	/**
	 * Mappe le buffer data sur la section [offet, offset+size[ du fichier relié au
	 * canal fcout.
	 * 
	 * @throws IOException toutes les exceptions d'entrées/sorties qui peuvent se
	 *                     produire.
	 */
	private void map() throws IOException {
		data = fcout.map(MapMode.READ_WRITE, offset, size);
	}

	/**
	 * Remplit le buffer mappé data à partir des données lues sur le canal fcin. La
	 * méthode ne fait qu'une seule lecture sur le canal. Elle sera appelée
	 * plusieurs fois jusqu'à avoir rempli le buffer. Cette méthode doit mapper le
	 * buffer data. Ce buffer n'est pas encore alloué.
	 * 
	 * @param fcin canal où lire les données. Le type de ce canal est
	 *             ReadableByteChannel, car il pourra recevoir aussi bien (1) des
	 *             références de FileChannel (si la source est un fichier), (2) que
	 *             des références de SocketChannel (si la source est une connexion
	 *             réseau TCP).
	 * @return le nombre d'octets lus. -1 en cas d'erreur ou de déconnexion.
	 * @throws IOException toutes les exceptions d'entrées/sorties qui peuvent se
	 *                     produire.
	 */
	public int fill(final ReadableByteChannel fcin) throws IOException {
		if (data == null) {
			this.map();
		}
		if (data.position() > 0) {
			GEN.trace("Block::read() complément à partir de " + data.position() + " sur " + data.capacity());
		}
		int lu = fcin.read(data);
		GEN.trace("Block::read() lu " + lu);
		return lu;
	}

	/**
	 * Indique si le buffer data est plein.
	 * 
	 * @return false si le buffer n'est pas encore mappé ou s'il n'est pas encore
	 *         plein, true si le buffer est mappé et plein.
	 */
	public boolean hasRemaining() {
		return data != null && data.hasRemaining();
	}

	/**
	 * getter sur l'offset.
	 * 
	 * @return la valeur de l'offset.
	 */
	public long getOffset() {
		return offset;
	}

	/**
	 * getter sur la taille du bloc.
	 * 
	 * @return la taille du bloc.
	 */
	public int getSize() {
		return size;
	}

	/**
	 * Ne fait rien si le buffer data n'est pas mappé, et sinon force l'écriture des
	 * données du buffer mappé data, et remet ce buffer à null.
	 */
	public void close() {
		if (data != null) {
			data.force();
			data = null;
		}
	}
}
