//CHECKSTYLE:OFF
// CHECKSTYLE OFF car il veut corriger un bug qui n'en est pas un.
package tsp.csc4509.tcp;

import static tsp.csc4509.common.Log.COMM;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;


/**
 * Classe qui
 *  1. établit la connexion d'un client en mode synchrone; 
 *  2. fournit les méthodes d'échange de données sur un canal TCP connecté (par un client ou un serveur).
 * 
 * @author Eric Lallet.
 *
 */
public class TcpSocketSynchrone implements AutoCloseable {

	/**
	 * un canal TCP connecté (soit par la connexion coté client, soit par l'accept coté serveur).
	 */
	private SocketChannel rwChan = null;


	

	/**
	 * Constructeur utilisé par les serveurs pour créer une instance TcpSocket à partir d'une connexion obtenue suite à un accept.
	 * @param rwChan 
	 *              le canal déjà connecté suite à l'accept.
	 */
	public TcpSocketSynchrone(final SocketChannel rwChan) {

		COMM.trace("TcpSocket: constructeur à partir d'un canal créé par un accept de serveur");	


		this.rwChan = rwChan;
	}

	/**
	 * Constructeur utilisé par le client pour ouvrir une connexion vers un serveur, et créer un instance TcpSocket.
	 * @param serverHost
	 * 	                le nom de la machine où tourne le serveur.
	 * @param serverPort
	 *                  le numéro du port TCP utilisé par le serveur.
	 * @throws IOException
	 *                  toutes les exceptions d'entrées/sorties.
	 */
	public TcpSocketSynchrone(final String serverHost, final int serverPort) throws IOException {

		COMM.trace("{}", () ->  "TcpSocket: constructeur d'un client: " + serverHost + ":" + serverPort);	

		InetSocketAddress rcvAddress = new InetSocketAddress(InetAddress.getByName(serverHost), serverPort);
		this.rwChan = SocketChannel.open(rcvAddress);
	}

	/**
	 * Envoie les données contenues dans le buffer sur la connexion TCP.
	 * @param buffer
	 *              le buffer contenant les données à envoyer.
	 * @return
	 *              le nombre d'octets envoyés.
	 * @throws IOException
	 *                  toutes les exceptions d'entrées/sorties.
	 */
	public int sendBuffer(final ByteBuffer buffer) throws IOException {

		COMM.trace("TcpSocket::sendBuffer");	

		return (rwChan.write(buffer));
	}


	/**
	 * Envoie un objet instancié sérialisable sur la connexion TCP.
	 * @param s
	 *         référence sur l'objet à envoyer.
	 * @return
	 *         le nombre d'octets envoyés.
	 * @throws IOException
	 *          toutes les exceptions d'entrées/sorties.
	 */
	public int sendObject(final Serializable s) throws IOException {

		COMM.trace("TcpSocket::sendObject");	

		ByteArrayOutputStream bo = new ByteArrayOutputStream();
		int size;

		ObjectOutputStream oo = new ObjectOutputStream(bo);
		oo.writeObject(s);
		oo.close();
		size = bo.size();
		sendSize(size);

		ByteBuffer buffer = ByteBuffer.allocate(size);
		buffer.put(bo.toByteArray());
		bo.close();
		buffer.flip();

		return sendBuffer(buffer);
	}

	/**
	 * Envoie un entier sur la connexion TCP.
	 * @param size
	 *            la valeur de l'entier à envoyer.
	 * @return
	 *            le nombre d'octets envoyés.
	 * @throws IOException
	 *            toutes les exceptions d'entrées/sorties.
	 */
	public int sendSize(final int size) throws IOException {

		COMM.trace("{}", () -> "TcpSocket::sendSize: " + size);	


		ByteBuffer buffer = ByteBuffer.allocate(Integer.BYTES);
		buffer.putInt(size);
		buffer.flip();
		return sendBuffer(buffer);
	}

	/**
	 * Reçoit des données sur la connexion TCP jusqu'à la fermeture la connexion entrante ou le remplissage du buffer passé en paramètre.
	 * La méthode remplit le buffer à partir du premier octet de celui-ci, en écrasant éventuellement des données déjà présente.
	 * @param buffer
	 *             le buffer où stocker les données.
	 * @return
	 *             le nombre d'octets placés dans le buffer.
	 * @throws IOException
	 *            toutes les exceptions d'entrées/sorties.
	 */
	public int receiveBuffer(final ByteBuffer buffer) throws IOException {

		COMM.trace("TcpSocket:: receiveBuffer.");	


		buffer.clear();
		while (rwChan.read(buffer) > 0) {
			// rien à faire tout est déjà fait.
			//         return du read = -1 => connexion du distant fermée
			//         return du read = 0 => buffer rempli

		}

		return buffer.position();

	}

	/**
	 * Reçoit un entier sur la connexion TCP.
	 * @return
	 *         la valeur de l'entier reçu.
	 * @throws IOException
	 *             le nombre d'octets placés dans le buffer.
	 */
	public int receiveSize() throws IOException {

		COMM.trace("TcpSocket::receiveSize.");	

		ByteBuffer buffer = ByteBuffer.allocate(Integer.BYTES);
		receiveBuffer(buffer);
		buffer.flip();
		return buffer.getInt();
	}

	/**
	 * Reçoit un objet sérialisable sur la connexion TCP.
	 * @return
	 *        la référence sur l'objet reçu.
	 * @throws IOException
	 *        toutes les exceptions d'entrées/sorties
	 * @throws ClassNotFoundException
	 *        l'exception levée si l'objet reçu est d'une classe inconnue de notre programme.
	 */
	public Serializable receiveObject() throws IOException, ClassNotFoundException {

		COMM.trace("TcpSocket::receiveObject.");	


		Serializable res = null;
		int size = receiveSize();

		COMM.trace("{}", () -> "TcpSocket::receiveObject en cours:" + size);	

		ByteBuffer buffer = ByteBuffer.allocate(size);
		receiveBuffer(buffer);
		buffer.flip(); // cette ligne est là pour la forme, mais n'est pas utile pour ce code
		// car la méthode buffer.array() ci dessous n'utilise pas les curseurs.

		ByteArrayInputStream bi = new ByteArrayInputStream(buffer.array());
		ObjectInputStream oi = new ObjectInputStream(bi);
		res = (Serializable) oi.readObject();
		oi.close();
		bi.close();

		COMM.trace("{}", () -> "TcpSocket::receiveObject terminé:" + size);	

		return res;

	}

	@Override
	public void close() throws IOException {

		COMM.trace("TcpSocket::close.");	

		this.rwChan.close();
	}
	
	/**
	 * Getteur sur le rwChan.
	 * @return
	 *       l'attribut rwChan.
	 */
	public SocketChannel getRwChan() {
		return rwChan;
	}


}
